/* cks_datastructures.h - CKS Datastructures header file
 * Copyright (C) 2001-2005 CryptNET, V. Alex Brennen (VAB)
 *
 * This file is part of the CryptNET OpenPGP Public Key Server (cks).
 *
 * cks is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * cks is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
 
#ifndef CKS_DATASTRUCTURES
#define CKS_DATASTRUCTURES

#include "datastructures.h"


struct openPGP_keyring
{
	struct openPGP_pubkey *pubkeys;
	char *buffer;
	unsigned long buffer_idx;
	char *radix_data;
	char encoded_cksum[5];
};

struct openPGP_pubkey
{
	int key_status;
	unsigned char key_version;
	unsigned int  key_revoked;
	unsigned char revoked_reason;
	unsigned int  algo_id;
	char algo[4];
	char keyid[8];
	char keyid_t[9];
	char fkeyid_t[17];
	char fp[20];
	char fp_t[61];
	char fp_db[41];
	char size[5];
	/* special data is 0x99 +2 needed for gening fp of subkeys */
	char special_data[3];
	unsigned int  key_size;
	long creation_time;
	long expiration_time;
	struct user_id *ids;
	struct openPGP_subkey *subkeys;
	/* This is all old openPGP_pubkey stuff */
	char *buffer;
	char *radix_data;
	char *radix_key;
	unsigned long buffer_idx;
	char encoded_cksum[5];
	struct openPGP_packet *the_packet;
	struct openPGP_packet *packet_list;
	int keyserver_no_modify;
	int has_photo;
	unsigned char *img_data;
	unsigned long image_len;

	struct openPGP_pubkey *prev;
	struct openPGP_pubkey *next;
};

struct user_id
{
	unsigned long packet_length;
	char *packet_data;
	char *id_data;
	unsigned int id_len;
	unsigned int revoked;
	struct key_signature *signatures;
	struct openPGP_packet *the_packet;

	struct user_id *prev;
	struct user_id *next;
};

struct key_signature
{
	unsigned char sig_version;
	unsigned char sig_type;
	char key_id[8];
	unsigned int  lkeyid;
	unsigned int  creation_time;
	struct openPGP_packet *the_packet;

	struct key_signature *prev;
	struct key_signature *next;
};

struct openPGP_subkey
{
	unsigned char subkey_version;
	unsigned int  algo_id;
	char algo[4];
	char keyid[8];
	char keyid_t[9];
	char fp[20];
	long creation_time;
	long expiration_time;
	unsigned int  key_size;
	struct key_signature *binding_signatures;
	struct openPGP_packet *the_packet;

	struct openPGP_subkey *prev;
	struct openPGP_subkey *next;
};

struct openPGP_packet
{
	int packet_id;
	unsigned long len_bytes;
	unsigned char pkt_len_d[2];
	unsigned char the_len_bytes[5];
	unsigned long packet_length;
	char *packet_data;
	unsigned long full_packet_length;
	char *full_packet_data;
	unsigned int  header_length;

	struct openPGP_packet *prev;
	struct openPGP_packet *next;
};

struct openPGP_subpacket
{
	int subpacket_id;
	unsigned long len_bytes;
	unsigned long subpacket_length;
	char *subpacket_data;

	struct openPGP_subpacket *prev;
	struct openPGP_subpacket *next;
};

struct openPGP_attribute
{
	char *attr_data;
	struct openPGP_packet *the_packet;
};

struct cks_config
{
	char cks_version[6];
	char bind_ip[16];
	char bind_port[10];
	int db_type;
	char dbsrvr_ip[16];
	char dbsrvr_port[10];
	char dbsrvr_db[16];
	int listen_threads;
	int use_cmnt;
	char vrsn[101];
	char cmnt[201];
	char adm_email[101];
	char sync_email[101];
	/* Accept Version 2 Keys */
	int acpt_v2;
	/* Accept Version 3 Keys */
	int acpt_v3;
	/* Maximum Query Time (real seconds) */
	int max_qry;
	/* Maximum Number Of Key To Return */
	int max_ret;
	/* Max keys to return per ID */
	int key_ret;
	/* Max keys w/ duplicate ID to accept */
	int dup_acpt;
	/* Display Link To Search BigLumber */
	int biglumber;
	/* Attempt Off Network Fetch */
	int aonf;
	/* Accept Public Key Image Packets */
	int image_pkts;

	char pid_file[256];
	char err_log[201];
	char mail_err_log[201];
	char data_log[201];
};

struct cks_ctx
{
	struct cks_config *config;
	struct d_linked_list *cache;

	int exit_flag;
	int verbose;
};


int init_ctx(struct cks_ctx **);
int init_openPGP_keyring(struct openPGP_keyring **,long);
int init_openPGP_pubkey(struct openPGP_pubkey **,long);
int init_user_id(struct user_id **);
int init_key_signature(struct key_signature **);
int init_openPGP_subkey(struct openPGP_subkey **);
int init_openPGP_packet(struct openPGP_packet **);
int init_openPGP_subpacket(struct openPGP_subpacket **);

struct openPGP_packet *get_first_packet(struct openPGP_packet *);
struct openPGP_packet *get_last_packet(struct openPGP_packet *);
struct openPGP_subpacket *get_first_subpacket(struct openPGP_packet *);
struct openPGP_subpacket *get_last_subpacket(struct openPGP_packet *);
struct openPGP_pubkey *get_first_pubkey(struct openPGP_pubkey *);
struct openPGP_pubkey *get_last_pubkey(struct openPGP_pubkey *);
struct user_id *get_first_uid(struct user_id *);
struct user_id *get_last_uid(struct user_id *);
struct key_signature *get_first_sig(struct key_signature *);
struct key_signature *get_last_sig(struct key_signature *);
struct openPGP_subkey *get_first_subkey(struct openPGP_subkey *);
struct openPGP_subkey *get_last_subkey(struct openPGP_subkey *);

int add_packet(struct openPGP_pubkey **, struct openPGP_packet *);
int add_pubkey(struct openPGP_keyring **, struct openPGP_pubkey *);
int add_subkey(struct openPGP_pubkey **, struct openPGP_subkey *);
int add_uid(struct openPGP_pubkey **, struct user_id *);
int add_sig(struct user_id **, struct key_signature *);
int add_subkey_binding_sig(struct openPGP_subkey **, struct key_signature *);

int extract_uid(struct user_id **);
int extract_sig(struct key_signature **);
int extract_subkey(struct openPGP_subkey **);
struct servers_to_sync * extract_srvr(struct servers_to_sync **);

int init_srvr_to_sync(struct servers_to_sync **);
struct servers_to_sync * add_server(struct servers_to_sync **, struct servers_to_sync *);
struct servers_to_sync * get_first_server(struct servers_to_sync *);
struct servers_to_sync * get_last_server(struct servers_to_sync *);
int count_servers_to_sync(struct servers_to_sync *);

void free_userid(struct user_id *);
void free_context(struct cks_ctx *);
void free_servers_to_sync(struct servers_to_sync *);
void free_keyring(struct openPGP_keyring *);
void free_pubkey(struct openPGP_pubkey *);
void free_pubkey_debug(struct openPGP_pubkey *);
void free_packet(struct openPGP_packet *);

#endif
